import numpy as np
import quaternion
from pint import Quantity


def plane_ray_intersection(
    ray: np.ndarray,
    position: np.ndarray,
    plane_normal: np.ndarray | None = None,
    p0: np.ndarray | None = None,
) -> np.ndarray:
    """
    Parameters
    ----------
    ray
        A rank 2 array with one ray per row. All rays must have a common point given
        by `position`
    position
        A point through which all the rays go through
    plane_normal
        Normal to the plane we want to intersect
    p0
        A point on the plane we want to intersect
    """
    ray = np.atleast_2d(ray)

    if plane_normal is None:
        plane_normal = np.array([0, 0, 1])

    if p0 is None:
        p0 = np.array([0, 0, 0])

    # This matrix product reduces to the dot product of ray and normal when there
    # is only one ray.
    denominator = ray @ plane_normal
    t = -np.dot(plane_normal, position - p0) / denominator

    intersection = position + t[:, None] * ray
    return intersection.squeeze()


def sphere_ray_intersection(
    rays: np.ndarray,
    rays_origin: np.ndarray,
    sphere_center: np.ndarray,
    sphere_radius: float,
    normalize_rays: bool = True,
) -> np.ndarray:
    """
    Return the closest intersection of the ray with the sphere.

    Notes
    -----
    - Assumption: Ray origin is outside of the sphere
    - Assumption: Ray direction is pointing towards the sphere

    Parameters
    ----------
    rays
        Rank 2 array of dimensions (n_rays, 3) with the cartesian components
        of the direction vectors. Normalization will happen internally unless
        it is disabled.
    rays_origin
        Rank 1 array with the origin for all the rays
    sphere_center
        Rank 1 array with the position of the sphere center
    sphere_radius
        Radius of the sphere
    normalize_rays
        The ray sphere intersection implemented requires the ray direction to
        be normalized. The default is to normalize internally but you can
        save cycles if you promise they are already normalized by setting
        this parameter to `False`.
    """
    # Validate parameters
    if rays.ndim != 2:
        e_ = "The rays array should have rank 2"
        raise ValueError(e_)

    if rays.shape[1] != 3:
        e_ = "The second dimension of the rays array should be 3"
        raise ValueError(e_)

    if rays_origin.ndim != 1 or len(rays_origin) != 3:
        e_ = "Rays origin should be a 1d array of length 3"
        raise ValueError(e_)

    # Make sure rays are normalized
    if normalize_rays:
        # The transpose dance makes broadcasting work
        rays = (rays.T / np.sqrt((rays**2).sum(axis=1))).T

    # Calculate intersections
    r0 = rays_origin - sphere_center
    ray_dot_r0 = np.dot(rays, r0)
    t = -ray_dot_r0 - np.sqrt(
        ray_dot_r0**2 - np.dot(r0, r0) + sphere_radius * sphere_radius
    )

    intersection = t[:, np.newaxis] * rays + r0

    return intersection


def find_aligning_quaternion(
    Q: quaternion.quaternion, u: np.ndarray, v: np.ndarray
) -> quaternion.quaternion:
    """Find the shortest quaternion that aligns two vectors.

    Parameters
    ----------
    Q
        Quaternion describing the initial orientation
    u
        3-element indexable describing the starting vector
    v
        3-element indexable describing the ending vector
    """
    U = quaternion.quaternion(0, u[0], u[1], u[2])
    V = quaternion.quaternion(0, v[0], v[1], v[2])

    return ((U * V).absolute() * Q - V * Q * U).normalized()


def get_sensor_rays(
    sensor_coords: np.ndarray,
    position: np.ndarray,
    orientation: quaternion.quaternion,
    focal_distance: Quantity,
    sensor_width: Quantity,
    sensor_height: Quantity,
    use_negative_sensor_plane: bool = False,
) -> np.ndarray:
    r"""
    Calculate the vectors that join the back nodal point to the sensor corners.

    Parameters
    ----------
    sensor_coords
        Rank 2 array of dimensions (n_rays, 2) with (u, v) coordinates of each ray.
    position
        3 element indexable with the position of back nodal point in cartesian coords.
    orientation: quaternion.quaternion
        Quaternion describing the orientation of the camera
    sensor_width: float
        Width of the sensor must be in the same units as `position`
    sensor_height: float
        Height of the sensor must be in the same units as `position`
    use_negative_sensor_plane
        Whether to use the negative sensor plane to calculate the corner rays.
        Use when you are dealing with a truly raw image from a sensor which is
        physically inverted due to passing through the camera center before
        hitting the sensor. See below for more explanations.

    Notes
    -----
    - `focal_distance`, `sensor_width` and `sensor_height` MUST be given in the same
      units.

    Negative VS Positive sensor plane
    ---------------------------------
    On a pinhole camera model (as appropriate for many real cameras) the image
    on the sensor comes out inverted. For this reason people will usually
    define a positive plane placed on front of the camera sensor where the
    image is not inverted. The name positive and negative come from the olden
    time when we had camera rolls that recorded a negative of the image.

    If you are dealing with raw images from a sensor you need to draw the
    camera rays from the actual physical place where the happen i.e. the
    sensor/negative. If this is not done then the corners of the sensors will
    be mapped to the wrong places on the ground. Why this inversion happens can
    be visualized by looking at the following pinhole camera diagram:

               d'
                ──────────────────── a'
               /\                '/
              /  \             ' /   negative
             /   \            ' /
            /     \          ' /
           /       \        ' /
        c' ────────────── ───/
            `        \     '  b'
             ``      \    '
               ``     \  '
                 ``  \  '               b->b' ray not shown
                   `` \'                      for clarity
                    ``\  camera center
                     ' \`
                     ' \ `v             `v` indicates direction of rays
                    '   \  ``
                   v    v     ``
              b   '      \      ``
                ─' ──────\ ───────`/ c
               /'         \       /
              /'          \      /
             /'            \    /
            /'             \   /
           /'               \ /
           ──────────────────/  positive
          a                   d

    The ground is down and camera rays go from the negative through the camera
    center and keep on going until they hit the ground.
    """
    # NOTE: Position is irrelevant as demonstrated by the test and under
    # `test_sensor_rays`. But I like it in the equations below because it makes
    # things clearer physically speaking.

    # Construct camera sensor basis by rotating the ECEF coordinate reference system
    # using the conjugate/inverted orientation. The conjugate is required because
    # the `orientation` takes us from the image sensor to the ECEF crs. The conjugate
    # takes us from the ECEF crs to the image sensor, which is what we need.
    camera_frame = np.identity(3)  # ECEF (cartesian) crs
    camera_frame = quaternion.rotate_vectors(
        orientation.conjugate(), camera_frame, axis=0
    )

    # NOTE: these are row vectors.
    u_ = camera_frame[0, :] * sensor_width  # x-axis
    v_ = camera_frame[1, :] * sensor_height  # y-axis
    w = camera_frame[2, :]  # z-axis

    # If we need to use the negative sensor plane then the images are inverted, i.e.
    # there is a switch of signs on the rows and columns. This sign will be used
    # to keep track of it.
    sign = -1 if use_negative_sensor_plane else 1

    # if `use_negative_plane == True` then the image sensor center is behind the camera
    # center instead of in front.
    # INFO: We call sensor to whatever plane we are using. If that's the positive please
    # do remember that in physical reality there is no sensor there.
    image_sensor_center = position + w * sign * focal_distance

    # x, y go from -0.5, to 0.5 with 0 being the center of the camera and
    # (-0.5, 0.5) the top left corner as viewed from the back nodal point
    # (0.5, -0.5) the bottom right corner as viewed from the BNP.

    def in_frame_coord(uv_coords):
        x, y = uv_coords[:, 0], uv_coords[:, 1]
        return image_sensor_center - x[:, None] * u_ - y[:, None] * v_

    # The sign below is to make sure that the rays are pointing toward the ground.
    # In the positive case (sign=+1) we go from the ray origin (aka camera center)
    # towards the corners of the sensor. In the negative (sign=-1) we go from the
    # corners of the sensor towards the camera center.
    return sign * (in_frame_coord(sensor_coords) - position)


def get_sensor_corner_rays(
    position: np.ndarray,
    orientation: quaternion.quaternion,
    focal_distance: Quantity,
    sensor_width: Quantity,
    sensor_height: Quantity,
    use_negative_sensor_plane: bool = False,
) -> tuple[np.ndarray, np.ndarray]:
    """
    Calculate the vectors that join the back nodal point to the sensor corners.

    Parameters
    ----------
    position
        3 element indexable with the position of back nodal point in cartesian coords.
    orientation: quaternion.quaternion
        Quaternion describing the orientation of the camera
    sensor_width: float
        Width of the sensor must be in the same units as `position`
    sensor_height: float
        Height of the sensor must be in the same units as `position`
    use_negative_sensor_plane
        Whether to use the negative sensor plane to calculate the rays.
    """
    top_left_corner = [-0.5, 0.5]
    top_right_corner = [0.5, 0.5]
    bottom_right_corner = [0.5, -0.5]
    bottom_left_corner = [-0.5, -0.5]

    sensor_corners = np.vstack(
        [top_left_corner, top_right_corner, bottom_right_corner, bottom_left_corner]
    )

    return position, get_sensor_rays(
        sensor_corners,
        position,
        orientation,
        focal_distance,
        sensor_width,
        sensor_height,
        use_negative_sensor_plane,
    )



"""Functions to work with ellipsoids and ellipsoidal coordinates."""

from dataclasses import dataclass
from math import atan2

import numpy as np
from sympy import Matrix, cos, lambdify, sin, solve, sqrt, symbols

from .constants import EARTH_SEMIMAJOR_AXIS, EARTH_SEMIMINOR_AXIS

# Symbols
# 𝜑: latitude
# 𝜆: longitude
# h: radial coordinate
# a: semi-major axis e
# e: eccentricity


@dataclass
class Ellipsoid:
    """An ellipsoid usually used as an approximation to the shape of the Earth.

    Attributes
    ----------
    major_axis
    minor_axis
    """

    major_axis: float
    minor_axis: float

    @property
    def eccentricity(self) -> float:
        """Return ellipsoid eccentricity."""
        return np.sqrt((self.major_axis**2 - self.minor_axis**2) / self.major_axis**2)


Earth = Ellipsoid(EARTH_SEMIMAJOR_AXIS, EARTH_SEMIMINOR_AXIS)


def xyz_to_ellipsoid(
    x: float,
    y: float,
    z: float,
    ellipsoid: Ellipsoid,
    tol: float | None = 1e-12,
) -> tuple[float, float, float]:
    """
    Transforms cartesian coordinates (x, y, z) into geodetic (longitude, latitude,
    elevation) coordinates.

    Parameters
    ----------
    x (meter)
       First component of the cartesian target coordinates
    y (meter)
       Second component of the cartesian target coordinates
    z (meter)
       Third component of the cartesian target coordinates
    tol
       Minimal change between two successive values of lon for the optimization loop to
       converge

    Returns
    -------
    lat (degree)
       Latitude of the target on the ellispoid
    lon (degree)
       Longitude of the target on the ellipsoid
    h (meter)
       Elevation of the target.
    """
    a = ellipsoid.major_axis
    e = ellipsoid.eccentricity

    𝜆 = atan2(y, x)

    # initialize variables
    p = np.sqrt((x**2) + (y**2))
    𝜑_old = atan2(z, p * (1 - e**2))

    converged = False
    𝜑_new = 0.0  # Just here to stop Ruff from complaining about unboud vars
    h = 0.0  # Same as before
    while not converged:
        # update parameters
        N = a / (np.sqrt(1 - (e**2) * np.sin(𝜑_old) ** 2))
        h = p / np.cos(𝜑_old) - N
        𝜑_new = atan2(z, p * (1 - (N / (N + h)) * e**2))
        # convergence criteria
        converged = np.abs(𝜑_old - 𝜑_new) < tol
        # new step
        𝜑_old = 𝜑_new

    𝜑 = 𝜑_new

    return np.rad2deg(𝜑), np.rad2deg(𝜆), h


def xyz_to_geodetic(
    x: float,
    y: float,
    z: float,
    tol: float | None = 1e-12,
):
    """Turn ECEF coordinates into geodetic coordinates.

    Returns
    -------
    lat (degree)
       Latitude of the target on the ellispoid
    lon (degree)
       Longitude of the target on the ellipsoid
    h (meter)
       Elevation of the target.
    """
    return xyz_to_ellipsoid(x, y, z, Earth, tol)


def _get_coord_functions_and_jacobian():
    𝜑, 𝜆, h, a, e = symbols("𝜑, 𝜆, h, a, e")

    N = a / sqrt(1 - (e * sin(𝜑)) ** 2)
    x = (N + h) * cos(𝜑) * cos(𝜆)
    y = (N + h) * cos(𝜑) * sin(𝜆)
    z = ((1 - e**2) * N + h) * sin(𝜑)

    coords = Matrix([x, y, z])
    vars_ = Matrix([𝜑, 𝜆, h])

    jacobian = coords.jacobian(vars_)

    basis_vectors_𝜑 = lambdify([𝜑, 𝜆, h, a, e], jacobian[:, 0], modules="numpy")
    basis_vectors_𝜆 = lambdify([𝜑, 𝜆, h, a, e], jacobian[:, 1], modules="numpy")
    basis_vectors_h = lambdify([𝜑, 𝜆, h, a, e], jacobian[:, 2], modules="numpy")

    coords = lambdify([𝜑, 𝜆, h, a, e], coords)

    return coords, [basis_vectors_𝜑, basis_vectors_𝜆, basis_vectors_h]


_coords, _basis_vector = _get_coord_functions_and_jacobian()


def ellipsoidal_to_xyz(
    𝜑: np.ndarray, 𝜆: np.ndarray, h: np.ndarray, ellipsoid: Ellipsoid
) -> np.ndarray:
    """
    Transform ellipsoidal coordinates to cartesian coordinates.

    All angular units are expected in radians.

    Parameters
    ----------
    𝜑
        Latitude, that is, angle subtended between the point and equator.
    𝜆
        Longitude
    h
        Elevation over the ellipsoid.
    ellipsoid
        The ellipsoid we are dealing with
    """
    major_axis, ecc = ellipsoid.major_axis, ellipsoid.eccentricity
    return _coords(𝜑, 𝜆, h, major_axis, ecc).squeeze()


def geodetic_to_xyz(𝜑: np.ndarray, 𝜆: np.ndarray, h: np.ndarray) -> np.ndarray:
    """
    Transform Earth geodetic coordinates to ECEF coordinates

    All angular units are expected in radians.

    Parameters
    ----------
    𝜑
        Latitude, that is, angle subtended between the point and equator.
    𝜆
        Longitude
    h
        Elevation over the ellipsoid.
    """
    major_axis, ecc = Earth.major_axis, Earth.eccentricity
    return _coords(𝜑, 𝜆, h, major_axis, ecc).squeeze()


def basis_at_point(
    𝜑: float, 𝜆: float, h: float, ellipsoid: Ellipsoid
) -> list[Ellipsoid]:
    """
    Returns the basis function for ellipsoidal coordinates.

    All angular units are expected in radians.

    Parameters
    ----------
    𝜑
        Latitude, that is, angle subtended between the point and equator.
    𝜆
        Longitude
    h
        Elevation over the ellipsoid.
    ellipsoid
        The ellipsoid we are dealing with
    """
    major_axis, ecc = ellipsoid.major_axis, ellipsoid.eccentricity

    basis_vectors_𝜑 = _basis_vector[0](𝜑, 𝜆, h, major_axis, ecc).squeeze()
    basis_vectors_𝜑 = basis_vectors_𝜑 / np.sqrt(np.sum(basis_vectors_𝜑**2))

    basis_vectors_𝜆 = _basis_vector[1](𝜑, 𝜆, h, major_axis, ecc).squeeze()
    basis_vectors_𝜆 = basis_vectors_𝜆 / np.sqrt(np.sum(basis_vectors_𝜆**2))

    basis_vectors_h = _basis_vector[2](𝜑, 𝜆, h, major_axis, ecc).squeeze()
    basis_vectors_h = basis_vectors_h / np.sqrt(np.sum(basis_vectors_h**2))

    return [basis_vectors_𝜆, basis_vectors_𝜑, basis_vectors_h]


def _solve_ray_ellipsoid_intersection():
    # Symbols
    # a: semimajor axis of the ellipsoid
    # b: semiminor axis of the ellipsoid
    # t: parametrizes position along the line
    # u,v,w: parametrize the ray direction
    # x0,y0,z0: parametrize the ray origin
    a, b, t, u, v, w, x0, y0, z0 = symbols("a, b, t, u, v, w, x0, y0, z0")

    ray_origin = Matrix([x0, y0, z0])
    ray_direction = Matrix([u, v, w])

    line = ray_origin + t * ray_direction

    # We want to solve the following equation == 0
    ts = solve((line[0] / a) ** 2 + (line[1] / a) ** 2 + (line[2] / b) ** 2 - 1, t)

    ts_first = lambdify([a, b, u, v, w, x0, y0, z0], ts[0], modules="numpy")
    ts_second = lambdify([a, b, u, v, w, x0, y0, z0], ts[1], modules="numpy")

    return ts_first, ts_second


_cached_ray_ellipsoid_solution = _solve_ray_ellipsoid_intersection()


def ray_ellipsoid_intersection(
    ray_origin: np.ndarray, ray_direction: np.ndarray, ellipsoid: Ellipsoid
) -> np.ndarray:
    """Calculate the intersection of a ray and an ellipsoid

    Parameters
    ----------
    ray_origin
        Origin point of ray
    ray_direction
        Direction of ray
    ellipsoid
        Ellipsoid to calculate intersection with

    Returns
    -------
        Intersection of a ray and ellipsoid
    """
    t_1 = _cached_ray_ellipsoid_solution[0](
        ellipsoid.major_axis,
        ellipsoid.minor_axis,
        *ray_direction,
        *ray_origin,
    )
    t_2 = _cached_ray_ellipsoid_solution[1](
        ellipsoid.major_axis,
        ellipsoid.minor_axis,
        *ray_direction,
        *ray_origin,
    )

    t_sol = min(t_1, t_2)

    return ray_origin + t_sol * ray_direction


def ray_Earth_intersection(
    ray_origin: np.ndarray, ray_direction: np.ndarray
) -> np.ndarray:
    """Calculate where a ray intersects with the Earth"""
    return ray_ellipsoid_intersection(ray_origin, ray_direction, Earth)


def basis_at_geoid(𝜑: float, 𝜆: float, h: float):
    """Returns the basis function for Earth coordinates"""
    return basis_at_point(𝜑, 𝜆, h, Earth)



"""Some useful constants."""

# NOTE: We don't use pint here because it's not used evenly enough across the
# codebase. If you end up using them it is recommended you attach units to them
# immediately.

# Source: https://github.com/geospace-code/pymap3d/blob/main/src/pymap3d/ellipsoid.py
EARTH_SEMIMAJOR_AXIS = 6378137.0  # WGS84 Earth semimajor axis (meters)
EARTH_SEMIMINOR_AXIS = 6356752.31424518  # WGS84 Earth semiminor axis (meters)
